"""Tests for loading Arbin .res files."""

# SPDX-FileCopyrightText: 2019-2020 Christopher Kerr <chris.kerr@mykolab.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import sqlite3
import subprocess

import pytest

from galvani import res2sqlite


have_mdbtools = subprocess.call(["which", "mdb-export"], stdout=subprocess.DEVNULL) == 0


def test_res2sqlite_help():
    """Test running `res2sqlite --help`.

    This should work even when mdbtools is not installed.
    """
    help_output = subprocess.check_output(["res2sqlite", "--help"])
    assert b"Convert Arbin .res files to sqlite3 databases" in help_output


@pytest.mark.skipif(
    have_mdbtools, reason="This tests the failure when mdbtools is not installed"
)
def test_convert_Arbin_no_mdbtools(testdata_dir, tmpdir):
    """Checks that the conversion fails with an appropriate error message."""
    res_file = os.path.join(testdata_dir, "arbin1.res")
    sqlite_file = os.path.join(str(tmpdir), "arbin1.s3db")
    with pytest.raises(
        RuntimeError, match="Could not locate the `mdb-export` executable."
    ):
        res2sqlite.convert_arbin_to_sqlite(res_file, sqlite_file)


@pytest.mark.skipif(
    not have_mdbtools, reason="Reading the Arbin file requires MDBTools"
)
@pytest.mark.parametrize("basename", ["arbin1", "UM34_Test005E"])
def test_convert_Arbin_to_sqlite_function(testdata_dir, tmpdir, basename):
    """Convert an Arbin file to SQLite using the functional interface."""
    res_file = os.path.join(testdata_dir, basename + ".res")
    sqlite_file = os.path.join(str(tmpdir), basename + ".s3db")
    res2sqlite.convert_arbin_to_sqlite(res_file, sqlite_file)
    assert os.path.isfile(sqlite_file)
    with sqlite3.connect(sqlite_file) as conn:
        csr = conn.execute("SELECT * FROM Channel_Normal_Table;")
        csr.fetchone()


@pytest.mark.skipif(
    not have_mdbtools, reason="Reading the Arbin file requires MDBTools"
)
def test_convert_cmdline(testdata_dir, tmpdir):
    """Checks that the conversion fails with an appropriate error message."""
    res_file = os.path.join(testdata_dir, "arbin1.res")
    sqlite_file = os.path.join(str(tmpdir), "arbin1.s3db")
    subprocess.check_call(["res2sqlite", res_file, sqlite_file])
    assert os.path.isfile(sqlite_file)
    with sqlite3.connect(sqlite_file) as conn:
        csr = conn.execute("SELECT * FROM Channel_Normal_Table;")
        csr.fetchone()



# -*- coding: utf-8 -*-
"""Code to read in data files from Bio-Logic instruments"""

# SPDX-FileCopyrightText: 2013-2020 Christopher Kerr, "bcolsen"
#
# SPDX-License-Identifier: GPL-3.0-or-later

__all__ = ["MPTfileCSV", "MPTfile"]

import re
import csv
from os import SEEK_SET
import time
from datetime import date, datetime, timedelta
from collections import defaultdict, OrderedDict

import numpy as np


def fieldname_to_dtype(fieldname):
    """Converts a column header from the MPT file into a tuple of
    canonical name and appropriate numpy dtype"""

    if fieldname == "mode":
        return ("mode", np.uint8)
    elif fieldname in (
        "ox/red",
        "error",
        "control changes",
        "Ns changes",
        "counter inc.",
    ):
        return (fieldname, np.bool_)
    elif fieldname in (
        "time/s",
        "P/W",
        "(Q-Qo)/mA.h",
        "x",
        "control/V",
        "control/mA",
        "control/V/mA",
        "(Q-Qo)/C",
        "dQ/C",
        "freq/Hz",
        "|Ewe|/V",
        "|I|/A",
        "Phase(Z)/deg",
        "|Z|/Ohm",
        "Re(Z)/Ohm",
        "-Im(Z)/Ohm",
        "Re(M)",
        "Im(M)",
        "|M|",
        "Re(Permittivity)",
        "Im(Permittivity)",
        "|Permittivity|",
        "Tan(Delta)",
    ):
        return (fieldname, np.float_)
    elif fieldname in (
        "Q charge/discharge/mA.h",
        "step time/s",
        "Q charge/mA.h",
        "Q discharge/mA.h",
        "Temperature/°C",
        "Efficiency/%",
        "Capacity/mA.h",
    ):
        return (fieldname, np.float_)
    elif fieldname in ("cycle number", "I Range", "Ns", "half cycle", "z cycle"):
        return (fieldname, np.int_)
    elif fieldname in ("dq/mA.h", "dQ/mA.h"):
        return ("dQ/mA.h", np.float_)
    elif fieldname in ("I/mA", "<I>/mA"):
        return ("I/mA", np.float_)
    elif fieldname in ("Ewe/V", "<Ewe>/V", "Ecell/V", "<Ewe/V>"):
        return ("Ewe/V", np.float_)
    elif fieldname.endswith(
        (
            "/s",
            "/Hz",
            "/deg",
            "/W",
            "/mW",
            "/W.h",
            "/mW.h",
            "/A",
            "/mA",
            "/A.h",
            "/mA.h",
            "/V",
            "/mV",
            "/F",
            "/mF",
            "/uF",
            "/µF",
            "/nF",
            "/C",
            "/Ohm",
            "/Ohm-1",
            "/Ohm.cm",
            "/mS/cm",
            "/%",
        )
    ):
        return (fieldname, np.float_)
    else:
        raise ValueError("Invalid column header: %s" % fieldname)


def comma_converter(float_text):
    """Convert text to float whether the decimal point is '.' or ','"""
    trans_table = bytes.maketrans(b",", b".")
    return float(float_text.translate(trans_table))


def MPTfile(file_or_path, encoding="ascii"):
    """Opens .mpt files as numpy record arrays

    Checks for the correct headings, skips any comments and returns a
    numpy record array object and a list of comments
    """

    if isinstance(file_or_path, str):
        mpt_file = open(file_or_path, "rb")
    else:
        mpt_file = file_or_path

    magic = next(mpt_file)
    if magic not in (b"EC-Lab ASCII FILE\r\n", b"BT-Lab ASCII FILE\r\n"):
        raise ValueError("Bad first line for EC-Lab file: '%s'" % magic)

    nb_headers_match = re.match(rb"Nb header lines : (\d+)\s*$", next(mpt_file))
    nb_headers = int(nb_headers_match.group(1))
    if nb_headers < 3:
        raise ValueError("Too few header lines: %d" % nb_headers)

    # The 'magic number' line, the 'Nb headers' line and the column headers
    # make three lines. Every additional line is a comment line.
    comments = [next(mpt_file) for i in range(nb_headers - 3)]

    fieldnames = next(mpt_file).decode(encoding).strip().split("\t")
    record_type = np.dtype(list(map(fieldname_to_dtype, fieldnames)))

    # Must be able to parse files where commas are used for decimal points
    converter_dict = dict(((i, comma_converter) for i in range(len(fieldnames))))
    mpt_array = np.loadtxt(mpt_file, dtype=record_type, converters=converter_dict)

    return mpt_array, comments


def MPTfileCSV(file_or_path):
    """Simple function to open MPT files as csv.DictReader objects

    Checks for the correct headings, skips any comments and returns a
    csv.DictReader object and a list of comments
    """

    if isinstance(file_or_path, str):
        mpt_file = open(file_or_path, "r")
    else:
        mpt_file = file_or_path

    magic = next(mpt_file)
    if magic.rstrip() != "EC-Lab ASCII FILE":
        raise ValueError("Bad first line for EC-Lab file: '%s'" % magic)

    nb_headers_match = re.match(r"Nb header lines : (\d+)\s*$", next(mpt_file))
    nb_headers = int(nb_headers_match.group(1))
    if nb_headers < 3:
        raise ValueError("Too few header lines: %d" % nb_headers)

    # The 'magic number' line, the 'Nb headers' line and the column headers
    # make three lines. Every additional line is a comment line.
    comments = [next(mpt_file) for i in range(nb_headers - 3)]

    mpt_csv = csv.DictReader(mpt_file, dialect="excel-tab")

    expected_fieldnames = (
        [
            "mode",
            "ox/red",
            "error",
            "control changes",
            "Ns changes",
            "counter inc.",
            "time/s",
            "control/V/mA",
            "Ewe/V",
            "dq/mA.h",
            "P/W",
            "<I>/mA",
            "(Q-Qo)/mA.h",
            "x",
        ],
        [
            "mode",
            "ox/red",
            "error",
            "control changes",
            "Ns changes",
            "counter inc.",
            "time/s",
            "control/V",
            "Ewe/V",
            "dq/mA.h",
            "<I>/mA",
            "(Q-Qo)/mA.h",
            "x",
        ],
        [
            "mode",
            "ox/red",
            "error",
            "control changes",
            "Ns changes",
            "counter inc.",
            "time/s",
            "control/V",
            "Ewe/V",
            "I/mA",
            "dQ/mA.h",
            "P/W",
        ],
        [
            "mode",
            "ox/red",
            "error",
            "control changes",
            "Ns changes",
            "counter inc.",
            "time/s",
            "control/V",
            "Ewe/V",
            "<I>/mA",
            "dQ/mA.h",
            "P/W",
        ],
    )
    if mpt_csv.fieldnames not in expected_fieldnames:
        raise ValueError("Unrecognised headers for MPT file format")

    return mpt_csv, comments


VMPmodule_hdr_v1 = np.dtype(
    [
        ("shortname", "S10"),
        ("longname", "S25"),
        ("length", "<u4"),
        ("version", "<u4"),
        ("date", "S8"),
    ]
)

VMPmodule_hdr_v2 = np.dtype(
    [
        ("shortname", "S10"),
        ("longname", "S25"),
        ("max length", "<u4"),
        ("length", "<u4"),
        ("version", "<u4"),
        ("unknown2", "<u4"),  # 10 for set, log and loop, 11 for data
        ("date", "S8"),
    ]
)

# Maps from colID to a tuple defining a numpy dtype
VMPdata_colID_dtype_map = {
    4: ("time/s", "<f8"),
    5: ("control/V/mA", "<f4"),
    6: ("Ewe/V", "<f4"),
    7: ("dq/mA.h", "<f8"),
    8: ("I/mA", "<f4"),  # 8 is either I or <I> ??
    9: ("Ece/V", "<f4"),
    11: ("<I>/mA", "<f8"),
    13: ("(Q-Qo)/mA.h", "<f8"),
    16: ("Analog IN 1/V", "<f4"),
    17: ("Analog IN 2/V", "<f4"),  # Probably column 18 is Analog IN 3/V, if anyone hits this error in the future
    19: ("control/V", "<f4"),
    20: ("control/mA", "<f4"),
    23: ("dQ/mA.h", "<f8"),  # Same as 7?
    24: ("cycle number", "<f8"),
    26: ("Rapp/Ohm", "<f4"),
    27: ("Ewe-Ece/V", "<f4"),
    32: ("freq/Hz", "<f4"),
    33: ("|Ewe|/V", "<f4"),
    34: ("|I|/A", "<f4"),
    35: ("Phase(Z)/deg", "<f4"),
    36: ("|Z|/Ohm", "<f4"),
    37: ("Re(Z)/Ohm", "<f4"),
    38: ("-Im(Z)/Ohm", "<f4"),
    39: ("I Range", "<u2"),
    69: ("R/Ohm", "<f4"),
    70: ("P/W", "<f4"),
    74: ("|Energy|/W.h", "<f8"),
    75: ("Analog OUT/V", "<f4"),
    76: ("<I>/mA", "<f4"),
    77: ("<Ewe>/V", "<f4"),
    78: ("Cs-2/µF-2", "<f4"),
    96: ("|Ece|/V", "<f4"),
    98: ("Phase(Zce)/deg", "<f4"),
    99: ("|Zce|/Ohm", "<f4"),
    100: ("Re(Zce)/Ohm", "<f4"),
    101: ("-Im(Zce)/Ohm", "<f4"),
    123: ("Energy charge/W.h", "<f8"),
    124: ("Energy discharge/W.h", "<f8"),
    125: ("Capacitance charge/µF", "<f8"),
    126: ("Capacitance discharge/µF", "<f8"),
    131: ("Ns", "<u2"),
    163: ("|Estack|/V", "<f4"),
    168: ("Rcmp/Ohm", "<f4"),
    169: ("Cs/µF", "<f4"),
    172: ("Cp/µF", "<f4"),
    173: ("Cp-2/µF-2", "<f4"),
    174: ("<Ewe>/V", "<f4"),
    178: ("(Q-Qo)/C", "<f4"),
    179: ("dQ/C", "<f4"),
    211: ("Q charge/discharge/mA.h", "<f8"),
    212: ("half cycle", "<u4"),
    213: ("z cycle", "<u4"),
    217: ("THD Ewe/%", "<f4"),
    218: ("THD I/%", "<f4"),
    220: ("NSD Ewe/%", "<f4"),
    221: ("NSD I/%", "<f4"),
    223: ("NSR Ewe/%", "<f4"),
    224: ("NSR I/%", "<f4"),
    230: ("|Ewe h2|/V", "<f4"),
    231: ("|Ewe h3|/V", "<f4"),
    232: ("|Ewe h4|/V", "<f4"),
    233: ("|Ewe h5|/V", "<f4"),
    234: ("|Ewe h6|/V", "<f4"),
    235: ("|Ewe h7|/V", "<f4"),
    236: ("|I h2|/A", "<f4"),
    237: ("|I h3|/A", "<f4"),
    238: ("|I h4|/A", "<f4"),
    239: ("|I h5|/A", "<f4"),
    240: ("|I h6|/A", "<f4"),
    241: ("|I h7|/A", "<f4"),
    242: ("|E2|/V", "<f4"),
    271: ("Phase(Z1) / deg", "<f4"),
    272: ("Phase(Z2) / deg", "<f4"),
    301: ("|Z1|/Ohm", "<f4"),
    302: ("|Z2|/Ohm", "<f4"),
    331: ("Re(Z1)/Ohm", "<f4"),
    332: ("Re(Z2)/Ohm", "<f4"),
    361: ("-Im(Z1)/Ohm", "<f4"),
    362: ("-Im(Z2)/Ohm", "<f4"),
    391: ("<E1>/V", "<f4"),
    392: ("<E2>/V", "<f4"),
    422: ("Phase(Zstack)/deg", "<f4"),
    423: ("|Zstack|/Ohm", "<f4"),
    424: ("Re(Zstack)/Ohm", "<f4"),
    425: ("-Im(Zstack)/Ohm", "<f4"),
    426: ("<Estack>/V", "<f4"),
    430: ("Phase(Zwe-ce)/deg", "<f4"),
    431: ("|Zwe-ce|/Ohm", "<f4"),
    432: ("Re(Zwe-ce)/Ohm", "<f4"),
    433: ("-Im(Zwe-ce)/Ohm", "<f4"),
    434: ("(Q-Qo)/C", "<f4"),
    435: ("dQ/C", "<f4"),
    438: ("step time/s", "<f8"),
    441: ("<Ecv>/V", "<f4"),
    462: ("Temperature/°C", "<f4"),
    467: ("Q charge/discharge/mA.h", "<f8"),
    468: ("half cycle", "<u4"),
    469: ("z cycle", "<u4"),
    471: ("<Ece>/V", "<f4"),
    473: ("THD Ewe/%", "<f4"),
    474: ("THD I/%", "<f4"),
    476: ("NSD Ewe/%", "<f4"),
    477: ("NSD I/%", "<f4"),
    479: ("NSR Ewe/%", "<f4"),
    480: ("NSR I/%", "<f4"),
    486: ("|Ewe h2|/V", "<f4"),
    487: ("|Ewe h3|/V", "<f4"),
    488: ("|Ewe h4|/V", "<f4"),
    489: ("|Ewe h5|/V", "<f4"),
    490: ("|Ewe h6|/V", "<f4"),
    491: ("|Ewe h7|/V", "<f4"),
    492: ("|I h2|/A", "<f4"),
    493: ("|I h3|/A", "<f4"),
    494: ("|I h4|/A", "<f4"),
    495: ("|I h5|/A", "<f4"),
    496: ("|I h6|/A", "<f4"),
    497: ("|I h7|/A", "<f4"),
    498: ("Q charge/mA.h", "<f8"),
    499: ("Q discharge/mA.h", "<f8"),
    500: ("step time/s", "<f8"),
    501: ("Efficiency/%", "<f8"),
    502: ("Capacity/mA.h", "<f8"),
    505: ("Rdc/Ohm", "<f4"),
    509: ("Acir/Dcir Control", "<u1"),
}

# These column IDs define flags which are all stored packed in a single byte
# The values in the map are (name, bitmask, dtype)
VMPdata_colID_flag_map = {
    1: ("mode", 0x03, np.uint8),
    2: ("ox/red", 0x04, np.bool_),
    3: ("error", 0x08, np.bool_),
    21: ("control changes", 0x10, np.bool_),
    31: ("Ns changes", 0x20, np.bool_),
    65: ("counter inc.", 0x80, np.bool_),
}


def parse_BioLogic_date(date_text):
    """Parse a date from one of the various formats used by Bio-Logic files."""
    date_formats = ["%m/%d/%y", "%m-%d-%y", "%m.%d.%y"]
    if isinstance(date_text, bytes):
        date_string = date_text.decode("ascii")
    else:
        date_string = date_text
    for date_format in date_formats:
        try:
            tm = time.strptime(date_string, date_format)
        except ValueError:
            continue
        else:
            break
    else:
        raise ValueError(
            f"Could not parse timestamp {date_string!r}"
            f" with any of the formats {date_formats}"
        )
    return date(tm.tm_year, tm.tm_mon, tm.tm_mday)


def VMPdata_dtype_from_colIDs(colIDs):
    """Get a numpy record type from a list of column ID numbers.

    The binary layout of the data in the MPR file is described by the sequence
    of column ID numbers in the file header. This function converts that
    sequence into a numpy dtype which can then be used to load data from the
    file with np.frombuffer().

    Some column IDs refer to small values which are packed into a single byte.
    The second return value is a dict describing the bit masks with which to
    extract these columns from the flags byte.

    """
    type_list = []
    field_name_counts = defaultdict(int)
    flags_dict = OrderedDict()
    for colID in colIDs:
        if colID in VMPdata_colID_flag_map:
            # Some column IDs represent boolean flags or small integers
            # These are all packed into a single 'flags' byte whose position
            # in the overall record is determined by the position of the first
            # column ID of flag type. If there are several flags present,
            # there is still only one 'flags' int
            if "flags" not in field_name_counts:
                type_list.append(("flags", "u1"))
                field_name_counts["flags"] = 1
            flag_name, flag_mask, flag_type = VMPdata_colID_flag_map[colID]
            # TODO what happens if a flag colID has already been seen
            # i.e. if flag_name is already present in flags_dict?
            # Does it create a second 'flags' byte in the record?
            flags_dict[flag_name] = (np.uint8(flag_mask), flag_type)
        elif colID in VMPdata_colID_dtype_map:
            field_name, field_type = VMPdata_colID_dtype_map[colID]
            field_name_counts[field_name] += 1
            count = field_name_counts[field_name]
            if count > 1:
                unique_field_name = "%s %d" % (field_name, count)
            else:
                unique_field_name = field_name
            type_list.append((unique_field_name, field_type))
        else:
            raise NotImplementedError(
                "Column ID {cid} after column {prev} "
                "is unknown".format(cid=colID, prev=type_list[-1][0])
            )
    return np.dtype(type_list), flags_dict


def read_VMP_modules(fileobj, read_module_data=True):
    """Reads in module headers in the VMPmodule_hdr format. Yields a dict with
    the headers and offset for each module.

    N.B. the offset yielded is the offset to the start of the data i.e. after
    the end of the header. The data runs from (offset) to (offset+length)"""
    while True:
        module_magic = fileobj.read(len(b"MODULE"))
        if len(module_magic) == 0:  # end of file
            break
        elif module_magic != b"MODULE":
            raise ValueError(
                "Found %r, expecting start of new VMP MODULE" % module_magic
            )
        VMPmodule_hdr = VMPmodule_hdr_v1

        # Reading headers binary information
        hdr_bytes = fileobj.read(VMPmodule_hdr.itemsize)
        if len(hdr_bytes) < VMPmodule_hdr.itemsize:
            raise IOError("Unexpected end of file while reading module header")

        # Checking if EC-Lab version is >= 11.50
        if hdr_bytes[35:39] == b"\xff\xff\xff\xff":
            VMPmodule_hdr = VMPmodule_hdr_v2
            hdr_bytes += fileobj.read(VMPmodule_hdr_v2.itemsize - VMPmodule_hdr_v1.itemsize)

        hdr = np.frombuffer(hdr_bytes, dtype=VMPmodule_hdr, count=1)
        hdr_dict = dict(((n, hdr[n][0]) for n in VMPmodule_hdr.names))
        hdr_dict["offset"] = fileobj.tell()
        if read_module_data:
            hdr_dict["data"] = fileobj.read(hdr_dict["length"])
            if len(hdr_dict["data"]) != hdr_dict["length"]:
                raise IOError(
                    """Unexpected end of file while reading data
                    current module: %s
                    length read: %d
                    length expected: %d"""
                    % (
                        hdr_dict["longname"],
                        len(hdr_dict["data"]),
                        hdr_dict["length"],
                    )
                )
            yield hdr_dict
        else:
            yield hdr_dict
            fileobj.seek(hdr_dict["offset"] + hdr_dict["length"], SEEK_SET)


MPR_MAGIC = b"BIO-LOGIC MODULAR FILE\x1a".ljust(48) + b"\x00\x00\x00\x00"


class MPRfile:
    """Bio-Logic .mpr file

    The file format is not specified anywhere and has therefore been reverse
    engineered. Not all the fields are known.

    Attributes
    ==========
    modules - A list of dicts containing basic information about the 'modules'
              of which the file is composed.
    data - numpy record array of type VMPdata_dtype containing the main data
           array of the file.
    startdate - The date when the experiment started
    enddate - The date when the experiment finished
    """

    def __init__(self, file_or_path):
        self.loop_index = None
        if isinstance(file_or_path, str):
            mpr_file = open(file_or_path, "rb")
        else:
            mpr_file = file_or_path
        magic = mpr_file.read(len(MPR_MAGIC))
        if magic != MPR_MAGIC:
            raise ValueError("Invalid magic for .mpr file: %s" % magic)

        modules = list(read_VMP_modules(mpr_file))

        self.modules = modules
        (settings_mod,) = (m for m in modules if m["shortname"] == b"VMP Set   ")
        (data_module,) = (m for m in modules if m["shortname"] == b"VMP data  ")
        maybe_loop_module = [m for m in modules if m["shortname"] == b"VMP loop  "]
        maybe_log_module = [m for m in modules if m["shortname"] == b"VMP LOG   "]

        n_data_points = np.frombuffer(data_module["data"][:4], dtype="<u4")
        n_columns = np.frombuffer(data_module["data"][4:5], dtype="u1").item()

        if data_module["version"] == 0:
            # If EC-Lab version >= 11.50, column_types is [0 1 0 3 0 174...] instead of [1 3 174...]
            if np.frombuffer(data_module["data"][5:6], dtype="u1").item():
                column_types = np.frombuffer(data_module["data"][5:], dtype="u1", count=n_columns)
                remaining_headers = data_module["data"][5 + n_columns:100]
                main_data = data_module["data"][100:]
            else:
                column_types = np.frombuffer(
                    data_module["data"][5:], dtype="u1", count=n_columns * 2
                )
                column_types = column_types[1::2]  # suppressing zeros in column types array
                # remaining headers should be empty except for bytes 5 + n_columns * 2
                # and 1006 which are sometimes == 1
                remaining_headers = data_module["data"][6 + n_columns * 2:1006]
                main_data = data_module["data"][1007:]
        elif data_module["version"] in [2, 3]:
            column_types = np.frombuffer(data_module["data"][5:], dtype="<u2", count=n_columns)
            # There are bytes of data before the main array starts
            if data_module["version"] == 3:
                num_bytes_before = 406  # version 3 added `\x01` to the start
            else:
                num_bytes_before = 405
            remaining_headers = data_module["data"][5 + 2 * n_columns:405]
            main_data = data_module["data"][num_bytes_before:]
        else:
            raise ValueError(
                "Unrecognised version for data module: %d" % data_module["version"]
            )

        assert not any(remaining_headers)

        self.dtype, self.flags_dict = VMPdata_dtype_from_colIDs(column_types)
        self.data = np.frombuffer(main_data, dtype=self.dtype)
        assert self.data.shape[0] == n_data_points

        # No idea what these 'column types' mean or even if they are actually
        # column types at all
        self.version = int(data_module["version"])
        self.cols = column_types
        self.npts = n_data_points
        self.startdate = parse_BioLogic_date(settings_mod["date"])

        if maybe_loop_module:
            (loop_module,) = maybe_loop_module
            if loop_module["version"] == 0:
                self.loop_index = np.frombuffer(loop_module["data"][4:], dtype="<u4")
                self.loop_index = np.trim_zeros(self.loop_index, "b")
            else:
                raise ValueError(
                    "Unrecognised version for data module: %d" % data_module["version"]
                )

        if maybe_log_module:
            (log_module,) = maybe_log_module
            self.enddate = parse_BioLogic_date(log_module["date"])

            # There is a timestamp at either 465 or 469 bytes
            # I can't find any reason why it is one or the other in any
            # given file
            ole_timestamp1 = np.frombuffer(
                log_module["data"][465:], dtype="<f8", count=1
            )
            ole_timestamp2 = np.frombuffer(
                log_module["data"][469:], dtype="<f8", count=1
            )
            ole_timestamp3 = np.frombuffer(
                log_module["data"][473:], dtype="<f8", count=1
            )
            ole_timestamp4 = np.frombuffer(
                log_module["data"][585:], dtype="<f8", count=1
            )

            if ole_timestamp1 > 40000 and ole_timestamp1 < 50000:
                ole_timestamp = ole_timestamp1
            elif ole_timestamp2 > 40000 and ole_timestamp2 < 50000:
                ole_timestamp = ole_timestamp2
            elif ole_timestamp3 > 40000 and ole_timestamp3 < 50000:
                ole_timestamp = ole_timestamp3
            elif ole_timestamp4 > 40000 and ole_timestamp4 < 50000:
                ole_timestamp = ole_timestamp4

            else:
                raise ValueError("Could not find timestamp in the LOG module")

            ole_base = datetime(1899, 12, 30, tzinfo=None)
            ole_timedelta = timedelta(days=ole_timestamp[0])
            self.timestamp = ole_base + ole_timedelta
            if self.startdate != self.timestamp.date():
                raise ValueError(
                    "Date mismatch:\n"
                    + "    Start date: %s\n" % self.startdate
                    + "    End date: %s\n" % self.enddate
                    + "    Timestamp: %s\n" % self.timestamp
                )

    def get_flag(self, flagname):
        if flagname in self.flags_dict:
            mask, dtype = self.flags_dict[flagname]
            return np.array(self.data["flags"] & mask, dtype=dtype)
        else:
            raise AttributeError("Flag '%s' not present" % flagname)



"""Helpers for pytest tests."""

# SPDX-FileCopyrightText: 2019 Christopher Kerr <chris.kerr@mykolab.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os

import pytest


@pytest.fixture(scope="session")
def testdata_dir():
    """Path to the testdata directory."""
    return os.path.join(os.path.dirname(__file__), "testdata")



#!/usr/bin/python

# SPDX-FileCopyrightText: 2013-2020 Christopher Kerr <chris.kerr@mykolab.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import subprocess as sp
import sqlite3
import re
import csv
import argparse
from copy import copy


# The following scripts are adapted from the result of running
# $ mdb-schema <result.res> oracle

mdb_tables = [
    "Version_Table",
    "Global_Table",
    "Resume_Table",
    "Channel_Normal_Table",
    "Channel_Statistic_Table",
    "Auxiliary_Table",
    "Event_Table",
    "Smart_Battery_Info_Table",
    "Smart_Battery_Data_Table",
]
mdb_5_23_tables = [
    "MCell_Aci_Data_Table",
    "Aux_Global_Data_Table",
    "Smart_Battery_Clock_Stretch_Table",
]
mdb_5_26_tables = [
    "Can_BMS_Info_Table",
    "Can_BMS_Data_Table",
]

mdb_tables_text = {
    "Version_Table",
    "Global_Table",
    "Event_Table",
    "Smart_Battery_Info_Table",
    "Can_BMS_Info_Table",
}
mdb_tables_numeric = {
    "Resume_Table",
    "Channel_Normal_Table",
    "Channel_Statistic_Table",
    "Auxiliary_Table",
    "Smart_Battery_Data_Table",
    "MCell_Aci_Data_Table",
    "Aux_Global_Data_Table",
    "Smart_Battery_Clock_Stretch_Table",
    "Can_BMS_Data_Table",
}

mdb_create_scripts = {
    "Version_Table": """
CREATE TABLE Version_Table
 (
    Version_Schema_Field    TEXT,
    Version_Comments_Field  TEXT
); """,
    "Global_Table": """
CREATE TABLE Global_Table
 (
    Test_ID                         INTEGER PRIMARY KEY,
    Test_Name                       TEXT,
    Channel_Index                   INTEGER,
    Start_DateTime                  REAL,
    DAQ_Index                       INTEGER,
    Channel_Type                    INTEGER,
    Creator                         TEXT,
    Comments                        TEXT,
    Schedule_File_Name              TEXT,
    Channel_Number                  INTEGER,
    Mapped_Aux_Voltage_Number       INTEGER,
    Mapped_Aux_Temperature_Number   INTEGER,
    Mapped_Aux_Pressure_Number      INTEGER,
    Mapped_Aux_PH_Number            INTEGER,
    Mapped_Aux_Flow_Rate_CNumber    INTEGER,
    Applications_Path               TEXT,
    Log_ChanStat_Data_Flag          INTEGER,
    Log_Aux_Data_Flag               INTEGER,
    Log_Event_Data_Flag             INTEGER,
    Log_Smart_Battery_Data_Flag     INTEGER,
  -- The following items are in 5.26 but not in 5.23
    Log_Can_BMS_Data_Flag           INTEGER DEFAULT NULL,
    Software_Version                TEXT DEFAULT NULL,
    Serial_Number                   TEXT DEFAULT NULL,
    Schedule_Version                TEXT DEFAULT NULL,
    MASS                            REAL DEFAULT NULL,
    Specific_Capacity               REAL DEFAULT NULL,
    Capacity                        REAL DEFAULT NULL,
  -- Item_ID exists in all versions
    Item_ID                         TEXT,
  -- These items are in 5.26 and 5.23 but not in 1.14
    Mapped_Aux_Conc_CNumber         INTEGER DEFAULT NULL,
    Mapped_Aux_DI_CNumber           INTEGER DEFAULT NULL,
    Mapped_Aux_DO_CNumber           INTEGER DEFAULT NULL
); """,
    "Resume_Table": """
CREATE TABLE Resume_Table
 (
    Test_ID         INTEGER PRIMARY KEY REFERENCES Global_Table(Test_ID),
    Step_Index      INTEGER,
    Cycle_Index     INTEGER,
    Channel_Status  INTEGER,
    Test_Time       REAL,
    Step_Time       REAL,
    CCapacity       REAL,
    DCapacity       REAL,
    CEnergy         REAL,
    DEnergy         REAL,
    TC_Time1        REAL,
    TC_Time2        REAL,
    TC_Time3        REAL,
    TC_Time4        REAL,
    TC_CCapacity1   REAL,
    TC_CCapacity2   REAL,
    TC_DCapacity1   REAL,
    TC_DCapacity2   REAL,
    TC_CEnergy1     REAL,
    TC_CEnergy2     REAL,
    TC_DEnergy1     REAL,
    TC_DEnergy2     REAL,
    MV_Counter1     REAL,
    MV_Counter2     REAL,
    MV_Counter3     REAL,
    MV_Counter4     REAL,
  -- Version 1.14 ends here, version 5.23 continues
    Charge_Time     REAL DEFAULT NULL,
    Discharge_Time  REAL DEFAULT NULL
); """,
    "Channel_Normal_Table": """
CREATE TABLE Channel_Normal_Table
 (
    Test_ID                 INTEGER REFERENCES Global_Table(Test_ID),
    Data_Point              INTEGER,
    Test_Time               REAL,
    Step_Time               REAL,
    DateTime                REAL,
    Step_Index              INTEGER,
    Cycle_Index             INTEGER,
    Is_FC_Data              INTEGER,
    Current                 REAL,
    Voltage                 REAL,
    Charge_Capacity         REAL,
    Discharge_Capacity      REAL,
    Charge_Energy           REAL,
    Discharge_Energy        REAL,
    "dV/dt"                 REAL,
    Internal_Resistance     REAL,
    AC_Impedance            REAL,
    ACI_Phase_Angle         REAL,
    PRIMARY KEY (Test_ID, Data_Point)
); """,
    "Channel_Statistic_Table": """
CREATE TABLE Channel_Statistic_Table
 (
    Test_ID                 INTEGER,
    Data_Point              INTEGER,
    Vmax_On_Cycle           REAL,
  -- Version 1.14 ends here, version 5.23 continues
    Charge_Time             REAL DEFAULT NULL,
    Discharge_Time          REAL DEFAULT NULL,
    PRIMARY KEY (Test_ID, Data_Point),
    FOREIGN KEY (Test_ID, Data_Point)
        REFERENCES Channel_Normal_Table (Test_ID, Data_Point)
); """,
    "Auxiliary_Table": """
CREATE TABLE Auxiliary_Table
 (
    Test_ID                 INTEGER,
    Data_Point              INTEGER,
    Auxiliary_Index         INTEGER,
    Data_Type               INTEGER,
    X                       REAL,
    "dX/dt"                 REAL,
    PRIMARY KEY (Test_ID, Data_Point, Auxiliary_Index),
    FOREIGN KEY (Test_ID, Data_Point)
        REFERENCES Channel_Normal_Table (Test_ID, Data_Point)
); """,
    "Event_Table": """
CREATE TABLE Event_Table
 (
    Test_ID                 INTEGER REFERENCES Global_Table(Test_ID),
    DateTime                REAL,
    Test_Time               REAL,
    Event_Type              INTEGER,
    Event_Describe          TEXT
); """,
    "Smart_Battery_Info_Table": """
CREATE TABLE Smart_Battery_Info_Table
 (
    Test_ID                 INTEGER PRIMARY KEY REFERENCES Global_Table(Test_ID),
    ManufacturerDate        REAL,
    ManufacturerAccess      TEXT,
    SpecificationInfo       TEXT,
    FullChargeCapacity      REAL,
    ChargingCurrent         REAL,
    ChargingVoltage         REAL,
    DesignCapacity          REAL,
    DesignVoltage           REAL,
    SerialNumber            INTEGER,
    ManufacturerName        TEXT,
    DeviceName              TEXT,
    DeviceChemistry         TEXT,
    ManufacturerData        TEXT,
  -- Version 1.14 ends here, version 5.23 continues
    Frequency               INTEGER DEFAULT NULL
); """,
    "Smart_Battery_Data_Table": """
CREATE TABLE Smart_Battery_Data_Table
 (
    Test_ID                 INTEGER,
    Data_Point              INTEGER,
    RemainingCapacityAlarm  REAL,
    RemainingTimeAlarm      INTEGER,
    BatteryMode             INTEGER,
    AtRate                  REAL,
    AtRateTimeToFull        INTEGER,
    AtRateTimeToEmpty       INTEGER,
    AtRateOK                INTEGER,
    Temperature             REAL,
    Voltage                 REAL,
    Current                 REAL,
    AverageCurrent          REAL,
    MaxError                INTEGER,
    RelativeStateOfCharge   INTEGER,
    AbsoluteStateOfCharge   INTEGER,
    RemainingCapacity       REAL,
    RunTimeToEmpty          INTEGER,
    AverageTimeToEmpty      INTEGER,
    AverageTimeToFull       INTEGER,
    BatteryStatus           INTEGER,
    CycleCount              INTEGER,
  -- Version 1.14 ends here, version 5.23 continues
    Pack_Status             INTEGER DEFAULT NULL,
    Pack_Configuration      INTEGER DEFAULT NULL,
    VCELL4                  REAL DEFAULT NULL,
    VCELL3                  REAL DEFAULT NULL,
    VCELL2                  REAL DEFAULT NULL,
    VCELL1                  REAL DEFAULT NULL,
    ManufacturerAccess      TEXT DEFAULT NULL,
    FullChargeCapacity      REAL DEFAULT NULL,
    BroadCast               INTEGER DEFAULT NULL,
    GPIO1                   INTEGER DEFAULT NULL,
    GPIO2                   INTEGER DEFAULT NULL,
    OptVCELL4               REAL DEFAULT NULL,
    OptVCELL3               REAL DEFAULT NULL,
    OptVCELL2               REAL DEFAULT NULL,
    OptVCELL1               REAL DEFAULT NULL,
    OMF1                    INTEGER DEFAULT NULL,
    OMF2                    INTEGER DEFAULT NULL,
    OMF3                    INTEGER DEFAULT NULL,
    OMF4                    INTEGER DEFAULT NULL,
    OMF5                    INTEGER DEFAULT NULL,
    FTEMP                   INTEGER DEFAULT NULL,
    STATUS                  INTEGER DEFAULT NULL,
    FET_TEMP                INTEGER DEFAULT NULL,
    ChargingCurrent         REAL DEFAULT NULL,
    ChargingVoltage         REAL DEFAULT NULL,
    ManufacturerData        REAL DEFAULT NULL,
  -- Version 5.23 ends here, version 5.26 continues
    BATMAN_Status           INTEGER DEFAULT NULL,
    DTM_PDM_Status          INTEGER DEFAULT NULL,
    PRIMARY KEY (Test_ID, Data_Point),
    FOREIGN KEY (Test_ID, Data_Point)
        REFERENCES Channel_Normal_Table (Test_ID, Data_Point)
); """,
    # The following tables are not present in version 1.14, but are in 5.23
    "MCell_Aci_Data_Table": """
CREATE TABLE MCell_Aci_Data_Table
 (
    Test_ID			INTEGER,
    Data_Point		INTEGER,
    Cell_Index		INTEGER,
    ACI			REAL,
    Phase_Shift		REAL,
    Voltage			REAL,
    Current			REAL,
    PRIMARY KEY (Test_ID, Data_Point, Cell_Index),
    FOREIGN KEY (Test_ID, Data_Point)
        REFERENCES Channel_Normal_Table (Test_ID, Data_Point)
);""",
    "Aux_Global_Data_Table": """
CREATE TABLE Aux_Global_Data_Table
 (
    Channel_Index		INTEGER,
    Auxiliary_Index	INTEGER,
    Data_Type		INTEGER,
    Nickname		TEXT,
    Unit			TEXT,
    PRIMARY KEY (Channel_Index, Auxiliary_Index, Data_Type)
);""",
    "Smart_Battery_Clock_Stretch_Table": """
CREATE TABLE Smart_Battery_Clock_Stretch_Table
 (
    Test_ID			    INTEGER,
    Data_Point		    INTEGER,
    ManufacturerAccess	    INTEGER,
    RemainingCapacityAlarm INTEGER,
    RemainingTimeAlarm	    INTEGER,
    BatteryMode		    INTEGER,
    AtRate			    INTEGER,
    AtRateTimeToFull	    INTEGER,
    AtRateTimeToEmpty	    INTEGER,
    AtRateOK		    INTEGER,
    Temperature		    INTEGER,
    Voltage			    INTEGER,
    Current			    INTEGER,
    AverageCurrent         INTEGER,
    MaxError		    INTEGER,
    RelativeStateOfCharge  INTEGER,
    AbsoluteStateOfCharge  INTEGER,
    RemainingCapacity	    INTEGER,
    FullChargeCapacity	    INTEGER,
    RunTimeToEmpty	    INTEGER,
    AverageTimeToEmpty	    INTEGER,
    AverageTimeToFull	    INTEGER,
    ChargingCurrent	    INTEGER,
    ChargingVoltage	    INTEGER,
    BatteryStatus		    INTEGER,
    CycleCount		    INTEGER,
    DesignCapacity	    INTEGER,
    DesignVoltage		    INTEGER,
    SpecificationInfo	    INTEGER,
    ManufacturerDate	    INTEGER,
    SerialNumber		    INTEGER,
    ManufacturerName	    INTEGER,
    DeviceName		    INTEGER,
    DeviceChemistry	    INTEGER,
    ManufacturerData	    INTEGER,
    Pack_Status		    INTEGER,
    Pack_Configuration	    INTEGER,
    VCELL4			    INTEGER,
    VCELL3			    INTEGER,
    VCELL2			    INTEGER,
    VCELL1			    INTEGER,
    PRIMARY KEY (Test_ID, Data_Point),
    FOREIGN KEY (Test_ID, Data_Point)
        REFERENCES Channel_Normal_Table (Test_ID, Data_Point)
);""",
    # The following tables are not present in version 5.23, but are in 5.26
    "Can_BMS_Info_Table": """
CREATE TABLE "Can_BMS_Info_Table"
 (
    Channel_Index        INTEGER PRIMARY KEY,
    CAN_Cfg_File_Name    TEXT,
    CAN_Configuration    TEXT
);
""",
    "Can_BMS_Data_Table": """
CREATE TABLE "Can_BMS_Data_Table"
 (
    Test_ID              INTEGER,
    Data_Point           INTEGER,
    CAN_MV_Index         INTEGER,
    Signal_Value_X       REAL,
    PRIMARY KEY (Test_ID, Data_Point, CAN_MV_Index),
    FOREIGN KEY (Test_ID, Data_Point)
        REFERENCES Channel_Normal_Table (Test_ID, Data_Point)
);
""",
}

mdb_create_indices = {
    "Channel_Normal_Table": """
CREATE UNIQUE INDEX data_point_index ON Channel_Normal_Table (Test_ID, Data_Point);
CREATE INDEX voltage_index ON Channel_Normal_Table (Test_ID, Voltage);
CREATE INDEX test_time_index ON Channel_Normal_Table (Test_ID, Test_Time);
"""
}

helper_table_script = """
CREATE TEMPORARY TABLE capacity_helper(
    Test_ID			    INTEGER NOT NULL,
    Cycle_Index		    INTEGER NOT NULL,
    Charge_Capacity        REAL NOT NULL,
    Discharge_Capacity     REAL NOT NULL,
    Charge_Energy          REAL NOT NULL,
    Discharge_Energy       REAL NOT NULL,
    FOREIGN KEY (Test_ID, Cycle_Index)
        REFERENCES Channel_Normal_Table (Test_ID, Cycle_Index)
);

INSERT INTO capacity_helper
    SELECT Test_ID, Cycle_Index,
           max(Charge_Capacity), max(Discharge_Capacity),
           max(Charge_Energy), max(Discharge_Energy)
        FROM Channel_Normal_Table
        GROUP BY Test_ID, Cycle_Index;

-- ## Alternative way of selecting ##
-- select *
--     from Channel_Normal_Table as a join Channel_Normal_Table as b
--         on (a.Test_ID = b.Test_ID and a.Data_Point = b.Data_Point + 1
--              and a.Charge_Capacity < b.Charge_Capacity);

DROP TABLE IF EXISTS Capacity_Sum_Table;
CREATE TABLE Capacity_Sum_Table(
    Test_ID			    INTEGER NOT NULL,
    Cycle_Index		    INTEGER NOT NULL,
    Charge_Capacity_Sum    REAL NOT NULL,
    Discharge_Capacity_Sum REAL NOT NULL,
    Charge_Energy_Sum      REAL NOT NULL,
    Discharge_Energy_Sum   REAL NOT NULL,
    FOREIGN KEY (Test_ID, Cycle_Index)
        REFERENCES Channel_Normal_Table (Test_ID, Cycle_Index)
);

INSERT INTO Capacity_Sum_Table
    SELECT a.Test_ID, a.Cycle_Index,
           total(b.Charge_Capacity), total(b.Discharge_Capacity),
           total(b.Charge_Energy), total(b.Discharge_Energy)
        FROM capacity_helper AS a LEFT JOIN capacity_helper AS b
            ON (a.Test_ID = b.Test_ID AND a.Cycle_Index > b.Cycle_Index)
        GROUP BY a.Test_ID, a.Cycle_Index;

DROP TABLE capacity_helper;

CREATE VIEW IF NOT EXISTS Capacity_View
    AS SELECT Test_ID, Data_Point, Test_Time, Step_Time, DateTime,
              Step_Index, Cycle_Index, Current, Voltage, "dV/dt",
              ( (Discharge_Capacity + Discharge_Capacity_Sum)
              - (Charge_Capacity + Charge_Capacity_Sum) ) AS Net_Capacity,
              ( (Discharge_Capacity + Discharge_Capacity_Sum)
              + (Charge_Capacity + Charge_Capacity_Sum) ) AS Gross_Capacity,
              ( (Discharge_Energy + Discharge_Energy_Sum)
              - (Charge_Energy + Charge_Energy_Sum) ) AS Net_Energy,
              ( (Discharge_Energy + Discharge_Energy_Sum)
              + (Charge_Energy + Charge_Energy_Sum) ) AS Gross_Energy
        FROM Channel_Normal_Table NATURAL JOIN Capacity_Sum_Table;
"""


def mdb_get_data_text(s3db, filename, table):
    print("Reading %s..." % table)
    insert_pattern = re.compile(
        r"""INSERT INTO "\w+" \([^)]+?\) VALUES (\((('[^']*')|"[^"]*"|[^')])+?\),?\s*)+;\n""",
        re.IGNORECASE,
    )
    try:
        # Initialize values to avoid NameError in except clause
        mdb_output = ""
        insert_match = None
        with sp.Popen(
            ["mdb-export", "-I", "postgres", filename, table],
            bufsize=-1,
            stdin=sp.DEVNULL,
            stdout=sp.PIPE,
            universal_newlines=True,
        ) as mdb_sql:
            mdb_output = mdb_sql.stdout.read()
            while len(mdb_output) > 0:
                insert_match = insert_pattern.match(mdb_output)
                s3db.execute(insert_match.group())
                mdb_output = mdb_output[insert_match.end():]
                mdb_output += mdb_sql.stdout.read()
            s3db.commit()

    except OSError as e:
        if e.errno == 2:
            raise RuntimeError(
                "Could not locate the `mdb-export` executable. "
                "Check that mdbtools is properly installed."
            )
        else:
            raise
    except BaseException:
        print("Error while importing %s" % table)
        if mdb_output:
            print("Remaining mdb-export output:", mdb_output)
        if insert_match:
            print("insert_re match:", insert_match)
        raise


def mdb_get_data_numeric(s3db, filename, table):
    print("Reading %s..." % table)
    try:
        with sp.Popen(
            ["mdb-export", filename, table],
            bufsize=-1,
            stdin=sp.DEVNULL,
            stdout=sp.PIPE,
            universal_newlines=True,
        ) as mdb_sql:
            mdb_csv = csv.reader(mdb_sql.stdout)
            mdb_headers = next(mdb_csv)
            quoted_headers = ['"%s"' % h for h in mdb_headers]
            joined_headers = ", ".join(quoted_headers)
            joined_placemarks = ", ".join(["?" for h in mdb_headers])
            insert_stmt = 'INSERT INTO "{0}" ({1}) VALUES ({2});'.format(
                table,
                joined_headers,
                joined_placemarks,
            )
            s3db.executemany(insert_stmt, mdb_csv)
            s3db.commit()
    except OSError as e:
        if e.errno == 2:
            raise RuntimeError(
                "Could not locate the `mdb-export` executable. "
                "Check that mdbtools is properly installed."
            )
        else:
            raise


def mdb_get_data(s3db, filename, table):
    if table in mdb_tables_text:
        mdb_get_data_text(s3db, filename, table)
    elif table in mdb_tables_numeric:
        mdb_get_data_numeric(s3db, filename, table)
    else:
        raise ValueError(
            "'%s' is in neither mdb_tables_text nor mdb_tables_numeric" % table
        )


def mdb_get_version(filename):
    """Get the version number from an Arbin .res file.

    Reads the Version_Table and parses the version from Version_Schema_Field.
    """
    print("Reading version number...")
    try:
        with sp.Popen(
            ["mdb-export", filename, "Version_Table"],
            bufsize=-1,
            stdin=sp.DEVNULL,
            stdout=sp.PIPE,
            universal_newlines=True,
        ) as mdb_sql:
            mdb_csv = csv.reader(mdb_sql.stdout)
            mdb_headers = next(mdb_csv)
            mdb_values = next(mdb_csv)
            try:
                next(mdb_csv)
            except StopIteration:
                pass
            else:
                raise ValueError(
                    "Version_Table of %s lists multiple versions" % filename
                )
    except OSError as e:
        if e.errno == 2:
            raise RuntimeError(
                "Could not locate the `mdb-export` executable. "
                "Check that mdbtools is properly installed."
            )
        else:
            raise
    if "Version_Schema_Field" not in mdb_headers:
        raise ValueError(
            "Version_Table of %s does not contain a Version_Schema_Field column"
            % filename
        )
    version_fields = dict(zip(mdb_headers, mdb_values))
    version_text = version_fields["Version_Schema_Field"]
    version_match = re.fullmatch("Results File ([.0-9]+)", version_text)
    if not version_match:
        raise ValueError(
            'File version "%s" did not match expected format' % version_text
        )
    version_string = version_match.group(1)
    version_tuple = tuple(map(int, version_string.split(".")))
    return version_tuple


def convert_arbin_to_sqlite(input_file, output_file):
    """Read data from an Arbin .res data file and write to a sqlite file.

    Any data currently in the sqlite file will be erased!
    """
    arbin_version = mdb_get_version(input_file)

    s3db = sqlite3.connect(output_file)

    tables_to_convert = copy(mdb_tables)
    if arbin_version >= (5, 23):
        tables_to_convert.extend(mdb_5_23_tables)
    if arbin_version >= (5, 26):
        tables_to_convert.extend(mdb_5_26_tables)

    for table in reversed(tables_to_convert):
        s3db.execute('DROP TABLE IF EXISTS "%s";' % table)

    for table in tables_to_convert:
        s3db.executescript(mdb_create_scripts[table])
        mdb_get_data(s3db, input_file, table)
        if table in mdb_create_indices:
            print("Creating indices for %s..." % table)
            s3db.executescript(mdb_create_indices[table])

    print("Creating helper table for capacity and energy totals...")
    s3db.executescript(helper_table_script)

    print("Vacuuming database...")
    s3db.executescript("VACUUM; ANALYZE;")


def main(argv=None):
    parser = argparse.ArgumentParser(
        description="Convert Arbin .res files to sqlite3 databases using mdb-export",
    )
    parser.add_argument("input_file", type=str)  # need file name to pass to sp.Popen
    parser.add_argument(
        "output_file", type=str
    )  # need file name to pass to sqlite3.connect

    args = parser.parse_args(argv)
    convert_arbin_to_sqlite(args.input_file, args.output_file)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2013-2020 Christopher Kerr <chris.kerr@mykolab.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os.path
import re
from datetime import date, datetime

import numpy as np
from numpy.testing import assert_array_almost_equal, assert_array_equal, assert_allclose
import pytest

from galvani import BioLogic, MPTfile, MPRfile
from galvani.BioLogic import MPTfileCSV  # not exported


def test_open_MPT(testdata_dir):
    mpt1, comments = MPTfile(os.path.join(testdata_dir, "bio_logic1.mpt"))
    assert comments == []
    assert mpt1.dtype.names == (
        "mode",
        "ox/red",
        "error",
        "control changes",
        "Ns changes",
        "counter inc.",
        "time/s",
        "control/V/mA",
        "Ewe/V",
        "dQ/mA.h",
        "P/W",
        "I/mA",
        "(Q-Qo)/mA.h",
        "x",
    )


def test_open_MPT_fails_for_bad_file(testdata_dir):
    with pytest.raises(ValueError, match="Bad first line"):
        MPTfile(os.path.join(testdata_dir, "bio_logic1.mpr"))


def test_open_MPT_csv(testdata_dir):
    mpt1, comments = MPTfileCSV(os.path.join(testdata_dir, "bio_logic1.mpt"))
    assert comments == []
    assert mpt1.fieldnames == [
        "mode",
        "ox/red",
        "error",
        "control changes",
        "Ns changes",
        "counter inc.",
        "time/s",
        "control/V/mA",
        "Ewe/V",
        "dq/mA.h",
        "P/W",
        "<I>/mA",
        "(Q-Qo)/mA.h",
        "x",
    ]


def test_open_MPT_csv_fails_for_bad_file(testdata_dir):
    with pytest.raises((ValueError, UnicodeDecodeError)):
        MPTfileCSV(os.path.join(testdata_dir, "bio_logic1.mpr"))


def test_colID_map_uniqueness():
    """Check some uniqueness properties of the VMPdata_colID_xyz maps."""
    field_colIDs = set(BioLogic.VMPdata_colID_dtype_map.keys())
    flag_colIDs = set(BioLogic.VMPdata_colID_flag_map.keys())
    field_names = [v[0] for v in BioLogic.VMPdata_colID_dtype_map.values()]
    flag_names = [v[0] for v in BioLogic.VMPdata_colID_flag_map.values()]
    assert not field_colIDs.intersection(flag_colIDs)
    # 'I/mA' and 'dQ/mA.h' are duplicated
    # assert len(set(field_names)) == len(field_names)
    assert len(set(flag_names)) == len(flag_names)
    assert not set(field_names).intersection(flag_names)


@pytest.mark.parametrize(
    "colIDs, expected",
    [
        ([1, 2, 3], [("flags", "u1")]),
        ([4, 6], [("time/s", "<f8"), ("Ewe/V", "<f4")]),
        ([1, 4, 21], [("flags", "u1"), ("time/s", "<f8")]),
        ([4, 6, 4], [("time/s", "<f8"), ("Ewe/V", "<f4"), ("time/s 2", "<f8")]),
        ([4, 9999], NotImplementedError),
    ],
)
def test_colID_to_dtype(colIDs, expected):
    """Test converting column ID to numpy dtype."""
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            BioLogic.VMPdata_dtype_from_colIDs(colIDs)
        return
    expected_dtype = np.dtype(expected)
    dtype, flags_dict = BioLogic.VMPdata_dtype_from_colIDs(colIDs)
    assert dtype == expected_dtype


@pytest.mark.parametrize(
    "data, expected",
    [
        ("02/23/17", date(2017, 2, 23)),
        ("10-03-05", date(2005, 10, 3)),
        ("11.12.20", date(2020, 11, 12)),
        (b"01/02/03", date(2003, 1, 2)),
        ("13.08.07", ValueError),
        ("03-04/05", ValueError),
    ],
)
def test_parse_BioLogic_date(data, expected):
    """Test the parse_BioLogic_date function."""
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            BioLogic.parse_BioLogic_date(data)
        return
    result = BioLogic.parse_BioLogic_date(data)
    assert result == expected


@pytest.mark.parametrize(
    "filename, startdate, enddate",
    [
        ("bio_logic1.mpr", "2011-10-29", "2011-10-31"),
        ("bio_logic2.mpr", "2012-09-27", "2012-09-27"),
        ("bio_logic3.mpr", "2013-03-27", "2013-03-27"),
        ("bio_logic4.mpr", "2011-11-01", "2011-11-02"),
        ("bio_logic5.mpr", "2013-01-28", "2013-01-28"),
        # bio_logic6.mpr has no end date because it does not have a VMP LOG module
        ("bio_logic6.mpr", "2012-09-11", None),
        # C019P-0ppb-A_C01.mpr stores the date in a different format
        ("C019P-0ppb-A_C01.mpr", "2019-03-14", "2019-03-14"),
        ("Rapp_Error.mpr", "2010-12-02", "2010-12-02"),
        ("Ewe_Error.mpr", "2021-11-18", "2021-11-19"),
        ("col_27_issue_74.mpr", "2022-07-28", "2022-07-28"),
    ],
)
def test_MPR_dates(testdata_dir, filename, startdate, enddate):
    """Check that the start and end dates in .mpr files are read correctly."""
    mpr = MPRfile(os.path.join(testdata_dir, filename))
    assert mpr.startdate.strftime("%Y-%m-%d") == startdate
    if enddate:
        assert mpr.enddate.strftime("%Y-%m-%d") == enddate
    else:
        assert not hasattr(mpr, "enddate")


def test_open_MPR_fails_for_bad_file(testdata_dir):
    with pytest.raises(ValueError, match="Invalid magic for .mpr file"):
        MPRfile(os.path.join(testdata_dir, "arbin1.res"))


def timestamp_from_comments(comments):
    for line in comments:
        time_match = re.match(b"Acquisition started on : ([0-9/]+ [0-9:]+)", line)
        if time_match:
            timestamp = datetime.strptime(
                time_match.group(1).decode("ascii"), "%m/%d/%Y %H:%M:%S"
            )
            return timestamp
    raise AttributeError("No timestamp in comments")


def assert_MPR_matches_MPT(mpr, mpt, comments):
    def assert_field_matches(fieldname, decimal):
        if fieldname in mpr.dtype.fields:
            assert_array_almost_equal(
                mpr.data[fieldname], mpt[fieldname], decimal=decimal
            )

    def assert_field_exact(fieldname):
        if fieldname in mpr.dtype.fields:
            assert_array_equal(mpr.data[fieldname], mpt[fieldname])

    assert_array_equal(mpr.get_flag("mode"), mpt["mode"])
    assert_array_equal(mpr.get_flag("ox/red"), mpt["ox/red"])
    assert_array_equal(mpr.get_flag("error"), mpt["error"])
    assert_array_equal(mpr.get_flag("control changes"), mpt["control changes"])
    if "Ns changes" in mpt.dtype.fields:
        assert_array_equal(mpr.get_flag("Ns changes"), mpt["Ns changes"])
    # Nothing uses the 0x40 bit of the flags
    assert_array_equal(mpr.get_flag("counter inc."), mpt["counter inc."])

    assert_array_almost_equal(
        mpr.data["time/s"], mpt["time/s"], decimal=2
    )  # 2 digits in CSV

    assert_field_matches("control/V/mA", decimal=6)
    assert_field_matches("control/V", decimal=6)

    assert_array_almost_equal(
        mpr.data["Ewe/V"], mpt["Ewe/V"], decimal=6
    )  # 32 bit float precision

    assert_field_matches("dQ/mA.h", decimal=16)  # 64 bit float precision
    assert_field_matches("P/W", decimal=10)  # 32 bit float precision for 1.xxE-5
    assert_field_matches("I/mA", decimal=6)  # 32 bit float precision

    assert_field_exact("cycle number")
    assert_field_matches("(Q-Qo)/C", decimal=6)  # 32 bit float precision

    try:
        assert timestamp_from_comments(comments) == mpr.timestamp.replace(microsecond=0)
    except AttributeError:
        pass


def assert_MPR_matches_MPT_v2(mpr, mpt, comments):
    """
    Asserts that the fields in the MPR.data ar the same as in the MPT.

    Modified from assert_MPR_matches_MPT. Automatically converts dtype from MPT data
    to dtype from MPR data before comparing the columns.

    Special case for EIS_indicators: these fields are valid only at f<100kHz so their
    values are replaced by -1 or 0 at high frequency in the MPT file, this is not the
    case in the MPR data.

    Parameters
    ----------
    mpr : MPRfile
        Data extracted with the MPRfile class.
    mpt : np.array
        Data extracted with MPTfile method.

    Returns
    -------
    None.

    """

    def assert_field_matches(fieldname):
        EIS_quality_indicators = [
            "THD Ewe/%",
            "NSD Ewe/%",
            "NSR Ewe/%",
            "|Ewe h2|/V",
            "|Ewe h3|/V",
            "|Ewe h4|/V",
            "|Ewe h5|/V",
            "|Ewe h6|/V",
            "|Ewe h7|/V",
            "THD I/%",
            "NSD I/%",
            "NSR I/%",
            "|I h2|/A",
            "|I h3|/A",
            "|I h4|/A",
            "|I h5|/A",
            "|I h6|/A",
            "|I h7|/A",
        ]

        if fieldname in EIS_quality_indicators:  # EIS quality indicators only valid for f < 100kHz
            index_inf_100k = np.where(mpr.data["freq/Hz"] < 100000)[0]
            assert_allclose(
                mpr.data[index_inf_100k][fieldname],
                mpt[index_inf_100k][fieldname].astype(mpr.data[fieldname].dtype),
            )
        elif fieldname == "<Ewe>/V":
            assert_allclose(
                mpr.data[fieldname],
                mpt["Ewe/V"].astype(mpr.data[fieldname].dtype),
            )
        elif fieldname == "<I>/mA":
            assert_allclose(
                mpr.data[fieldname],
                mpt["I/mA"].astype(mpr.data[fieldname].dtype),
            )
        elif fieldname == "dq/mA.h":
            assert_allclose(
                mpr.data[fieldname],
                mpt["dQ/mA.h"].astype(mpr.data[fieldname].dtype),
            )
        else:
            assert_allclose(
                mpr.data[fieldname],
                mpt[fieldname].astype(mpr.data[fieldname].dtype),
            )

    def assert_field_exact(fieldname):
        if fieldname in mpr.dtype.fields:
            assert_array_equal(mpr.data[fieldname], mpt[fieldname])

    for key in mpr.flags_dict.keys():
        assert_array_equal(mpr.get_flag(key), mpt[key])

    for d in mpr.dtype.descr[1:]:
        assert_field_matches(d[0])

    try:
        assert timestamp_from_comments(comments) == mpr.timestamp.replace(microsecond=0)
    except AttributeError:
        pass


@pytest.mark.parametrize(
    "basename",
    [
        "bio_logic1",
        "bio_logic2",
        # No bio_logic3.mpt file
        "bio_logic4",
        # bio_logic5 and bio_logic6 are special cases
        "CV_C01",
        "121_CA_455nm_6V_30min_C01",
        "020-formation_CB5",
    ],
)
def test_MPR_matches_MPT(testdata_dir, basename):
    """Check the MPR parser against the MPT parser.

    Load a binary .mpr file and a text .mpt file which should contain
    exactly the same data. Check that the loaded data actually match.
    """
    binpath = os.path.join(testdata_dir, basename + ".mpr")
    txtpath = os.path.join(testdata_dir, basename + ".mpt")
    mpr = MPRfile(binpath)
    mpt, comments = MPTfile(txtpath, encoding="latin1")
    assert_MPR_matches_MPT(mpr, mpt, comments)


def test_MPR5_matches_MPT5(testdata_dir):
    mpr = MPRfile(os.path.join(testdata_dir, "bio_logic5.mpr"))
    mpt, comments = MPTfile(
        (
            re.sub(b"\tXXX\t", b"\t0\t", line)
            for line in open(os.path.join(testdata_dir, "bio_logic5.mpt"), mode="rb")
        )
    )
    assert_MPR_matches_MPT(mpr, mpt, comments)


def test_MPR6_matches_MPT6(testdata_dir):
    mpr = MPRfile(os.path.join(testdata_dir, "bio_logic6.mpr"))
    mpt, comments = MPTfile(os.path.join(testdata_dir, "bio_logic6.mpt"))
    mpr.data = mpr.data[:958]  # .mpt file is incomplete
    assert_MPR_matches_MPT(mpr, mpt, comments)


@pytest.mark.parametrize(
    "basename_v1150",
    ["v1150_CA", "v1150_CP", "v1150_GCPL", "v1150_GEIS", "v1150_MB", "v1150_OCV", "v1150_PEIS"],
)
def test_MPR_matches_MPT_v1150(testdata_dir, basename_v1150):
    """Check the MPR parser against the MPT parser.

    Load a binary .mpr file and a text .mpt file which should contain
    exactly the same data. Check that the loaded data actually match.
    """
    binpath = os.path.join(testdata_dir, "v1150", basename_v1150 + ".mpr")
    txtpath = os.path.join(testdata_dir, "v1150", basename_v1150 + ".mpt")
    mpr = MPRfile(binpath)
    mpt, comments = MPTfile(txtpath, encoding="latin1")
    assert_MPR_matches_MPT_v2(mpr, mpt, comments)



# -*- coding: utf-8 -*-
# SPDX-FileCopyrightText: 2014-2020 Christopher Kerr <chris.kerr@mykolab.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os.path

from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), "README.md")) as f:
    readme = f.read()

setup(
    name="galvani",
    version="0.4.1",
    description="Open and process battery charger log data files",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/echemdata/galvani",
    author="Chris Kerr",
    author_email="chris.kerr@mykolab.ch",
    license="GPLv3+",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering :: Chemistry",
    ],
    packages=["galvani"],
    entry_points={
        "console_scripts": [
            "res2sqlite = galvani.res2sqlite:main",
        ],
    },
    python_requires=">=3.6",
    install_requires=["numpy"],
    tests_require=["pytest"],
)
